﻿/// @file parse.cpp
///
/// @brief パーサー
///
#include "parse.h"
#include <wx/textfile.h>
#include <wx/regex.h>
#include <wx/filename.h>
#include <wx/wfstream.h>
#include <wx/arrimpl.cpp>

#define DATA_DIR _T("data")
#define CHAR_CODE_TABLE _T("char_code.dat")
#define BASIC_CODE_TABLE _T("basic_code.dat")

#define VALS_SIZE	511

/// 改行コード
static wxString nl_chr[] = {
	_T("\r"), _T("\n"), _T("\r\n")
};
/// BOMコード
#define BOM_CODE "\xEF\xBB\xBF"
/// スタートアドレス
static int iStartAddr[4] = {
	0x0768,	// newon11 0x0400 + 0x368
	0x0B68,	// newon15 0x0800 + 0x368
	0x2768,	// newon 3 0x2400 + 0x368
	0x4768 	// newon 7 0x4400 + 0x368
};

///
/// パーサークラス
///
Parse::Parse(const wxString &app_path)
{
	mAppPath = app_path;
	mBigEndian = true;	// 6809なので

	mNewLineAscii = 0;
	mNewLineUtf8 = 1;
	mIncludeBOM = false;
	mStartAddr = 2;
}

Parse::~Parse()
{
}

bool Parse::Init()
{
	// 文字コード変換テーブルの読み込み
	if (LoadCharCodeTable() != psOK) {
		return false;
	}
	// BASICコード変換テーブルの読み込み
	if (LoadBasicCodeTable() != psOK) {
		return false;
	}
	return true;
}

// 指定したファイルを開く
bool Parse::OpenDataFile(const wxString &in_file_name)
{
	PsFileInfo in_file;
	in_file.file_name.Assign(in_file_name);
	if (!in_file.file.Open(in_file.file_name.GetFullPath(), in_file.file.read)) {
		mErrInfo.SetInfo(__LINE__, psError, psErrFileNotFound);
		mErrInfo.ShowMsgBox();
		return false;
	}
	if (!CheckDataFormat(in_file)) {
		return false;
	}
	in_file.file.Close();

	CloseDataFile();
	mInFile = in_file;
	mInFile.file.Open(mInFile.file_name.GetFullPath(), mInFile.file.read);

	return true;
}

// 入力データのフォーマットチェック
bool Parse::CheckDataFormat(PsFileInfo &in_file_info)
{
	wxUint8 hsign[3];
	bool st = true;

	size_t len = in_file_info.file.Read((void *)hsign, 3);

	if (len == 0) {
		// file is empty
		mErrInfo.SetInfo(__LINE__, psError, psErrFileIsEmpty);
		mErrInfo.ShowMsgBox();
		return false;
	}

	if (hsign[0] == 0xff) {
		// basic intermediate language
		in_file_info.start_pos = 3;
		st = CheckBinaryDataFormat(in_file_info);

	} else if (hsign[0] == 0xfe) {
		// basic intermediate language (encrypted)
		in_file_info.type.SetFlag(psAscii, false);
		in_file_info.type.SetFlag(psEncrypted, true);
		// error
		mErrInfo.SetInfo(__LINE__, psError, psErrFileIsEncrypted);
		mErrInfo.ShowMsgBox();
		return false;

	} else if (memcmp(hsign, BOM_CODE, 3) == 0) {
		// UTF-8 text with BOM?
		in_file_info.start_pos = 3;
		st = CheckAsciiDataFormat(in_file_info);
		in_file_info.type.SetFlag(psUTF8BOM, true);

	} else if (hsign[0] != 0x0a && hsign[0] != 0x0d && hsign[0] != 0x09 && hsign[0] < 0x20) {
		// ummm machine code???
		in_file_info.type.SetFlag(psAscii, false);
		// error
		mErrInfo.SetInfo(__LINE__, psError, psErrBasic);
		mErrInfo.ShowMsgBox();
		return false;

	} else {
		// propably text
		in_file_info.start_pos = 0;
		st = CheckAsciiDataFormat(in_file_info);
	}

	in_file_info.file.Seek(in_file_info.start_pos, wxFromStart);

	return st;
}

// 中間言語形式データのフォーマットチェック
bool Parse::CheckBinaryDataFormat(PsFileInfo &in_file_info)
{
	PsFileType     out_type;
	wxArrayString out_lines;
	ParseResult   result;
	wxArrayString basic_types;
	bool st = true;

	in_file_info.type.SetFlag(psAscii, false);

	// ファイルを解析する
	GetBasicTypes(basic_types);
	out_type.SetFlag(psAscii, true);
	for(size_t i=0; i < basic_types.GetCount(); i++) {
		in_file_info.file.Seek(in_file_info.start_pos, wxFromStart);
		in_file_info.type.SetBasicType(basic_types[i]);
		out_type.SetBasicType(basic_types[i]);
		out_lines.Empty();
		result.Empty();
		st = ReadBinaryToAscii(in_file_info, out_type, out_lines, &result);
		if (result.GetCount() > 0) {
			// 変換できないBASIC statementがある
			in_file_info.type.SetFlag(psExtendBasic, true);
			out_type.SetFlag(psExtendBasic, true);
		} else {
			break;
		}
	}
	if (st != true) {
		// エラー
		mErrInfo.SetInfo(__LINE__, psError, psErrParseBasic);
		mErrInfo.ShowMsgBox();
		return false;
	}

	mParsedData.Empty();
	Report(result, mParsedData);

	// UTF-8に変換
	mParsedType = out_type;
	mParsedType.SetFlag(psUTF8, true);
	mParsedType.SetCharType(GetCharType(0));
	st = ConvAsciiToUTF8(out_lines, mParsedType, &mParsedData);

	return true;
}

// アスキー形式データのフォーマットチェック
bool Parse::CheckAsciiDataFormat(PsFileInfo &in_file_info)
{
	PsFileType     out_type;
	wxArrayString tmp_lines;
	wxArrayString out_lines;
	ParseResult   result;
	bool st;

	out_lines.Empty();
	mParsedData.Empty();

	// まずUTF-8として変換できるかを試みる
	in_file_info.type.SetFlag(psAscii | psUTF8, true);
	in_file_info.file.Seek(in_file_info.start_pos, wxFromStart);
	st = ReadAsciiText(in_file_info, out_lines, true);
	if (st) {
		// UTF-8らしい
		wxArrayString char_types;
		GetCharTypes(char_types);
		in_file_info.type.SetFlag(psAscii | psUTF8, true);
		out_type.SetFlag(psAscii, true);
		for(size_t i=0; i<char_types.GetCount(); i++) {
			// 一旦Asciiにして変換できるかどうか調べる
			result.Empty();
			tmp_lines.Empty();
			in_file_info.type.SetCharType(char_types.Item(i));
			st = ConvUTF8ToAscii(in_file_info.type, out_lines, &tmp_lines, &result);
			if (result.GetCount() == 0) {
				// エラーなし
				break;
			}
		}
		if (st != true) {
			// エラー解析中止
			mErrInfo.SetInfo(__LINE__, psError, psErrInvalidString);
			mErrInfo.ShowMsgBox();
			return false;
		}
		mParsedType = in_file_info.type;
		Report(result, mParsedData);
		// UTF-8文字に変換する
		st = ConvAsciiToUTF8(tmp_lines, mParsedType, &mParsedData);
	} else {
		// UTF-8で読めない？アスキー形式
		in_file_info.type.SetFlag(psAscii, true);
		in_file_info.type.SetFlag(psUTF8, false);
		in_file_info.file.Seek(in_file_info.start_pos, wxFromStart);
		out_lines.Empty();
		mParsedData.Empty();
		st = ReadAsciiText(in_file_info, out_lines, false);
		// interace 文字に変換する
		mParsedType.SetFlag(psAscii | psUTF8, true);
		mParsedType.SetCharType(GetCharType(0));
		st = ConvAsciiToUTF8(out_lines, mParsedType, &mParsedData);
	}

	return true;
}

// 入力ファイルの形式を変更して読み直す
bool Parse::ReloadOpendAsciiData(int type, int mask, const wxString &char_type)
{
	if (!mInFile.file.IsOpened()) {
		// ファイル開いていない
		return false;
	}

	PsFileType out_type(&mParsedType);

	mInFile.type.SetFlag(mask, false);
	mInFile.type.SetFlag(type, true);
	mInFile.type.SetCharType(char_type);

	return ReloadAsciiData(mInFile, out_type);
}
// 表示用データの形式を変更して読み直す
bool Parse::ReloadParsedData(int type, int mask, const wxString &char_type)
{
	if (!mInFile.file.IsOpened()) {
		// ファイル開いていない
		return false;
	}

	PsFileType out_type(&mParsedType);

	out_type.SetFlag(mask, false);
	out_type.SetFlag(type, true);
	out_type.SetCharType(char_type);

	if (mInFile.type.GetFlag(psAscii)) {
		return ReloadAsciiData(mInFile, out_type);
	} else {
		return ReloadBinaryData(mInFile, out_type);
	}
}
// 中間言語形式データの読み直し
bool Parse::ReloadOpenedBinaryData(int type, int mask, const wxString &basic_type)
{
	if (!mInFile.file.IsOpened() || mInFile.type.GetFlag(psAscii)) {
		// ファイル開いていないかASCII形式の場合
		return false;
	}

	PsFileType out_type(&mParsedType);

	mInFile.type.SetFlag(mask, false);
	mInFile.type.SetFlag(type, true);
	mInFile.type.SetBasicType(basic_type);
	out_type.SetBasicType(basic_type);

	return ReloadBinaryData(mInFile, out_type);
}

// 中間言語形式データの読み直し
bool Parse::ReloadBinaryData(PsFileInfo &in_file_info, PsFileType &out_type)
{
	wxArrayString out_lines;
	ParseResult result;
	bool st;

	out_lines.Empty();
	mParsedData.Empty();
	result.Empty();

	in_file_info.file.Seek(in_file_info.start_pos, wxFromStart);
	st = ReadBinaryToAscii(in_file_info, out_type, out_lines, &result);
	Report(result, mParsedData);
	// interace or noninterace 文字に変換する
	mParsedType = out_type;
	st = ConvAsciiToUTF8(out_lines, mParsedType, &mParsedData);

	return true;
}

// アスキー形式データの読み直し
bool Parse::ReloadAsciiData(PsFileInfo &in_file_info, PsFileType &out_type)
{
	wxArrayString tmp_lines;
	wxArrayString in_lines;
	ParseResult result;
	bool st;

	tmp_lines.Empty();
	in_lines.Empty();
	mParsedData.Empty();
	result.Empty();

	//
	in_file_info.file.Seek(in_file_info.start_pos, wxFromStart);
	if (in_file_info.type.GetFlag(psUTF8)) {
		st = ReadAsciiText(in_file_info, in_lines, true);
		// 一旦ASCIIにする
		st = ConvUTF8ToAscii(in_file_info.type, in_lines, &tmp_lines, &result);
		Report(result, mParsedData);
		// interace or noninterace 文字に変換する
		mParsedType = out_type;
		st = ConvAsciiToUTF8(tmp_lines, mParsedType, &mParsedData);
	} else {
		st = ReadAsciiText(in_file_info, in_lines, false);
		// interace or noninterace 文字に変換する
		mParsedType = out_type;
		st = ConvAsciiToUTF8(in_lines, mParsedType, &mParsedData);
	}
	return true;
}

// ファイルを閉じる
void Parse::CloseDataFile()
{
	if (mInFile.file.IsOpened()) {
		mInFile.file.Close();
	}
}

// ファイル開いているか
bool Parse::IsOpenedDataFile()
{
	return mInFile.file.IsOpened();
}

// 開いているファイル種類
PsFileType *Parse::GetOpenedDataTypePtr()
{
	return &mInFile.type;
}

// 開いているファイル情報
PsFileInfo *Parse::GetOpenedDataInfoPtr()
{
	return &mInFile;
}

// ファイル名
wxString Parse::GetFileNameBase()
{
	wxString name = mInFile.file_name.GetName();
	return name;
}

// 出力ファイルを開く
bool Parse::OpenOutFile(const wxString &out_file_name, PsFileType &file_type)
{
	// 入力ファイルと同じファイルはダメ
	wxFileName out_file(out_file_name);
	if (mInFile.file_name == out_file) {
		mErrInfo.SetInfo(__LINE__, psError, pwErrSameFile);
		mErrInfo.ShowMsgBox();
		return false;
	}
	// 出力ファイルオープン
	mOutFile.file_name.Assign(out_file_name);
	mOutFile.type = file_type;
	mInFile.type.SetBasicType(file_type.GetBasicType());
	if (!mOutFile.file.Open(mOutFile.file_name.GetFullPath(), mInFile.file.write)) {
		mErrInfo.SetInfo(__LINE__, psError, pwErrCannotWrite);
		mErrInfo.ShowMsgBox();
		return false;
	}

	return true;
}

// 出力ファイルを閉じる
void Parse::CloseOutFile()
{
	if (mOutFile.file.IsOpened()) {
		mOutFile.file.Close();
	}
}

// エクスポート
bool Parse::ExportData()
{
	bool st = false;
	wxArrayString inlines;
	wxArrayString outlines;
	ParseResult result;

	inlines.Empty();
	outlines.Empty();
	result.Empty();

	mInFile.file.Seek(mInFile.start_pos, wxFromStart);

	// BOM付きで出力するか
	mOutFile.type.SetFlag(psUTF8BOM, mIncludeBOM);

	if (mInFile.type.GetFlag(psUTF8)) {
		if (mOutFile.type.GetFlag(psUTF8)) {
			// UTF-8テキストからUTF-8テキストに変換
			wxArrayString tmplines;
			st = ReadAsciiText(mInFile, inlines, true);
			st = ConvUTF8ToAscii(mInFile.type, inlines, &tmplines, &result);
			st = ConvAsciiToUTF8(tmplines, mOutFile.type, &outlines);
			st = WriteText(outlines, mOutFile);
		} else if (mOutFile.type.GetFlag(psAscii)) {
			// UTF-8テキストからアスキー形式に変換
			st = ReadAsciiText(mInFile, inlines, true);
			st = ConvUTF8ToAscii(mInFile.type, inlines, &outlines, &result);
			st = WriteText(outlines, mOutFile);
		} else {
			// UTF-8テキストから中間言語に変換
			st = ReadAsciiText(mInFile, inlines, true);
			st = ConvUTF8ToAscii(mInFile.type, inlines, &outlines, &result);
			st = ParseAsciiToBinary(mInFile.type, outlines, mOutFile, &result);
		}
	} else if (mInFile.type.GetFlag(psAscii)) {
		if (mOutFile.type.GetFlag(psUTF8)) {
			// アスキー形式からUTF-8テキストに変換
			st = ReadAsciiText(mInFile, inlines, false);
			st = ConvAsciiToUTF8(inlines, mOutFile.type, &outlines);
			st = WriteText(outlines, mOutFile);
		} else if (mOutFile.type.GetFlag(psAscii)) {
			// アスキー形式からアスキー形式に変換
			st = ReadAsciiText(mInFile, inlines, false);
			st = WriteText(inlines, mOutFile);
		} else {
			// アスキー形式から中間言語に変換
			st = ReadAsciiText(mInFile, inlines, false);
			st = ParseAsciiToBinary(mInFile.type, inlines, mOutFile, &result);
		}
	} else {
		// 中間言語からアスキー形式/UTF-8テキストに変換
		if (mOutFile.type.GetFlag(psAscii)) {
			st = ReadBinaryToAscii(mInFile, mOutFile.type, inlines, &result);
			if (mOutFile.type.GetFlag(psUTF8)) {
				// UTF-8テキストの場合
				st = ConvAsciiToUTF8(inlines, mOutFile.type, &outlines);
			} else {
				outlines = inlines;
			}
			st = WriteText(outlines, mOutFile);
		}
	}

	mParsedData.Empty();
	if (result.GetCount() > 0) {
		Report(result, mParsedData);
		st = false;
	} else {
		mParsedData.Add(_("----- Complete. -----"));
		mParsedData.Add(_T(""));
	}
	return st;
}

// 中間言語からアスキー形式テキストに変換
bool Parse::ReadBinaryToAscii(PsFileInfo &in_file, PsFileType &out_file_type, wxArrayString &outlines, ParseResult *result)
{
	bool extend_basic = out_file_type.GetFlag(psExtendBasic);	// DISK BASICか
	wxString basic_type = out_file_type.GetBasicType();

	wxString name = _("Binary->Ascii");

//	wxUint8 hsign[3];			// header
	long next_addr;				// next address
	long line_number;			// line number
	// body
	wxUint8 vals[10];
//	long vall;
//	bool command_extend = false;
	bool quoted = false;
	bool data_area = false;
	bool comment_area = false;

	wxString body = _T("");
	wxString chrstr;
	CodeMapItem *item;

	int vrow = 0;
	int vcol = 0;
	int vlen = 0;

	int phase = 1;

	int error_count = 0;	// エラー発生数
	bool stopped = false;

	// parse start
//	outlines.Add(_T(""));

	while(!in_file.file.Eof() && phase >= 0) {
		memset(vals, 0, sizeof(vals));
		if (error_count > 20) {
			// エラーが多いので解析中止
			stopped = true;
			break;
		}
		switch(phase) {
//		case 0:
//			// get header
//			in_file.file.Read((void *)hsign, 3);
//			// next phase
//			phase = 1;
//			break;
//
		case 1:
			// get next address
			in_file.file.Read((void *)vals, 2);
			next_addr = BytesToLong(vals, 2);
			if (next_addr == 0) {
				// end
				phase = -1;
			} else {
				// next phase
				phase = 2;
				vrow++;
			}
			break;

		case 2:
			// get line number
			in_file.file.Read((void *)vals, 2);
			line_number = BytesToLong(vals, 2);
			// next phase
			phase = 3;

			vcol = 0;
			break;

		case 3:
			// get 10 chars

			vlen = (int)in_file.file.Read((void *)vals, 10);

			if (vlen == 0) {
				// end
				phase = -1;
				break;
			} else {
				if (vals[0] == 0) {
					// end of line
					phase = 4;
					in_file.file.Seek(1-vlen, wxFromCurrent);
					vcol++;
					break;
				} else if (vals[0] == 0x22) {
					// double quote
					chrstr = wxString::From8BitData((const char *)vals, 1);
					body += chrstr;
					in_file.file.Seek(1-vlen, wxFromCurrent);
					vcol++;
					if (data_area != true && comment_area != true) {
						quoted = quoted ? false : true;	// toggle quoted string
					}
				} else {
					if (quoted || data_area || comment_area) {
						// quoted string or DATA line or REM line
						chrstr = wxString::From8BitData((const char *)vals, 1);
						body += chrstr;
						in_file.file.Seek(1-vlen, wxFromCurrent);
						vcol++;

					} else {
						// find command statement
						mBasicCodeTbl.FindSection(_T(""));
						item = mBasicCodeTbl.FindByCode(vals);
						if (item == NULL && extend_basic) {
							mBasicCodeTbl.FindSection(basic_type);
							item = mBasicCodeTbl.FindByCode(vals);
						}
						if (item != NULL) {
							if (item->FindAttr(_T("colon"))) { // ELSE statement
								if (body.Right(1) == _T(":")) {
									body = body.Left(body.Len()-1); // trim last char ':'
								}
							}

							body += item->GetStr();
							in_file.file.Seek((int)(item->GetCodeLength())-vlen, wxFromCurrent);
							vcol+=item->GetCodeLength();

							if (item->FindAttr(_T("data"))) { // DATA statement

								data_area = true;
							} else if (item->FindAttr(_T("comment"))) { // ' REM statement

								comment_area = true;
							}
						} else {
							chrstr = _T("");
							if (vals[0] < 0x20 || 0x80 <= vals[0]) {
								// Unknown command (error?)
								error_count++;
								// for parse result
								if (result) {
									result->Add(vrow, vcol, line_number, name, prErrInvalidBasicCode);
								}
							}
							chrstr = wxString::From8BitData((const char *)vals, 1);
							body += chrstr;
							in_file.file.Seek(1-vlen, wxFromCurrent);
							vcol++;
						}
					}
				}
			}
			break;

		case 4:
			// end of line
			chrstr.Printf(_T("%ld "),line_number);
			outlines.Add(chrstr + body);

			// clear val
			quoted = false;
			data_area = false;
			comment_area = false;
			body = _T("");

			// next phase
			phase = 1;
			break;

		}
	}
	if (stopped) {
		if (result) {
			result->Add(vrow, vcol, line_number, name, prErrStopInvalidBasicCode);
		}
		return false;
	}
	return true;

}

// アスキー形式/UTF-8テキストから中間言語に変換して出力
bool Parse::ParseAsciiToBinary(PsFileType &in_file_type, wxArrayString &inlines, PsFileInfo &out_file, ParseResult *result)
{
	bool rc = true;
	char buf[8];

	mNextAddress = iStartAddr[mStartAddr];

	wxString outline = _T("");

	// header
	if (out_file.file.IsOpened()) {
		sprintf(buf, "%c%c%c",0xff,0xff,0xff);
		out_file.file.Write((const void *)buf, 3);
	}

	// body
	for(size_t row = 0; row < inlines.GetCount(); row++) {
		rc = ParseAsciiToBinaryOneLine(in_file_type, inlines[row], out_file, row+1, result);
	}

	// footer
	if (out_file.file.IsOpened()) {
		sprintf(buf, "%c%c",0x00,0x00);
		out_file.file.Write((const void *)buf, 2);
	}

	return true;

}

// アスキー形式1行を中間言語に変換して出力
bool Parse::ParseAsciiToBinaryOneLine(PsFileType &in_file_type, wxString &in_data, PsFileInfo &out_file, int row, ParseResult *result)
{
	if (in_data.IsEmpty()) {
		return true;
	}

	bool extend_basic = in_file_type.GetFlag(psExtendBasic);	// DISK BASICか
	wxString basic_type = in_file_type.GetBasicType();

	wxString name = _("Ascii->Binary");

	bool quoted = false;
	bool data_area = false;
	bool comment_area = false;
	int  variable_area = 0;

	wxRegEx reAlphaNumeric(_T("^[0-9a-zA-Z]"));
	wxRegEx reAlpha(_T("^[a-zA-Z]"));

	wxString chrstr;
	bool is_empty_chr;
//	long     chr;
//	char     chrvals[4];
	CodeMapItem *item;

	long line_number;

	wxUint8 buf[8];

	wxUint8 body[VALS_SIZE+1];
	size_t  body_len = 0;

	size_t pos = 0;

	// line number
	line_number = GetLineNumber(in_data, &pos);

	// body
	while(pos < in_data.Len()) {

		if (in_data[pos] == wxChar('\"')) {
			// quote
			quoted = quoted ? false : true;
		}
		// variable
		chrstr.Empty();
		is_empty_chr = true;
		if (variable_area) {
			if (!reAlphaNumeric.Matches(in_data.Mid(pos,1)) || (variable_area > 16)) {
				// end of values area
				variable_area = 0;
			} else {
				variable_area++;

				chrstr = in_data[pos];
				chrstr = chrstr.Upper();
				is_empty_chr = false;

				body_len += AscStrToBytes(chrstr, &body[body_len], sizeof(body)-body_len);
				pos++;
			}
		}
		// find command
		if (comment_area != true
		 && data_area != true
		 && quoted != true
		 && variable_area == 0) {
			// search command
			mBasicCodeTbl.FindSection(_T(""));
			item = mBasicCodeTbl.FindByStr(in_data.Mid(pos).Upper(), true);
			if (item == NULL && extend_basic) {
				mBasicCodeTbl.FindSection(basic_type);
				item = mBasicCodeTbl.FindByStr(in_data.Mid(pos).Upper(), true);
			}
			if (item != NULL) {
				// find a BASIC sentence
				if (item->FindAttr(_T("colon"))) {
					body_len += AscStrToBytes(_T(":"), &body[body_len], sizeof(body)-body_len);
				}

				chrstr = wxString::From8BitData((const char *)item->GetCode(), item->GetCodeLength());
				is_empty_chr = false;

				if (item->FindAttr(_T("comment"))) {
					comment_area = true;
				}
				if (item->FindAttr(_T("data"))) {
					data_area = true;
				}

				body_len += AscStrToBytes(chrstr, &body[body_len], sizeof(body)-body_len);
				pos += item->GetStr().Len();
			} else {
				// non BASIC sentence
				if (reAlpha.Matches(in_data.Mid(pos,1))) {
					// is variable
					variable_area = 1;

					chrstr = in_data[pos];
					chrstr = chrstr.Upper();
					is_empty_chr = false;

					body_len += AscStrToBytes(chrstr, &body[body_len], sizeof(body)-body_len);
					pos++;
				} else if ((wxUint32)in_data[pos] >= 0x80) {
					// unknown char ?(error)
					if (result) result->Add(row+1,pos,line_number,name,prErrInvalidChar);
				}
			}
		}
		if (is_empty_chr) {
			chrstr = in_data[pos];
			body_len += AscStrToBytes(chrstr, &body[body_len], sizeof(body)-body_len);
			pos++;
		}
	}
	mNextAddress += body_len + 5;

	if (out_file.file.IsOpened()) {
		LongToBytes(mNextAddress, buf, 2);
		LongToBytes(line_number, &buf[2], 2);
		out_file.file.Write((const void *)buf, 4);
		out_file.file.Write((const void *)body, body_len);
		buf[0] = '\0';
		out_file.file.Write((const void *)buf, 1);
	}
	return true;
}

// アスキー形式テキストを読む
bool Parse::ReadAsciiText(PsFileInfo &in_file, wxArrayString &outlines, bool to_utf8)
{
	bool rc = true;	// utf8に変換できない場合 false

	wxUint8 vals[VALS_SIZE + 1];
	size_t  valpos = VALS_SIZE;
	size_t  valsta = VALS_SIZE;
	size_t  vallen = VALS_SIZE;
	size_t  valend = VALS_SIZE;
	vals[VALS_SIZE] = 0;

	wxString body;

	int lf_len = 0;

	int phase = 0;

	memset((void *)vals, 0, sizeof(vals));
//	inlines.Add(_T(""));

	while(phase >= 0) {
		switch(phase) {
		case 0:
			if (valsta > 0) {
				// shift
				for(size_t i=0; i<(vallen + 1 - valsta); i++)
					vals[i] = vals[i + valsta];
				// read
				size_t vlen = valend - valsta;
				vallen = in_file.file.Read((void *)&vals[vlen], VALS_SIZE - vlen);
				if (vallen + vlen < VALS_SIZE) {
					valend = vallen + vlen;
				} else {
					valend = VALS_SIZE;
				}
				vallen = VALS_SIZE;
				valpos -= valsta;
				valsta = 0;
			}
			phase = 1;
			break;

		case 1:
			if (valpos >= valend) {
				// end of line
				phase = 2;
				break;
			} else if (vals[valpos] == '\r' && vals[valpos+1] == '\n') {
				// 改行
				lf_len = 2;
				phase = 2;
				break;
			} else if (vals[valpos] == '\r' || vals[valpos] == '\n') {
				// 改行
				lf_len = 1;
				phase = 2;
				break;
			}

			valpos++;

			if (valpos + 32 > vallen) {
				// continue
				phase = 0;
			}
			break;

		case 2:
			// 1行出力
			vals[valpos] = 0;

			if (valsta < valpos) {	// 空行は省く
				if (to_utf8) {
					// utf8に変換を試みる
					body = wxString((const char *)&vals[valsta], wxConvUTF8);
					if (body.IsEmpty()) {
						// utf8に変換できない
						rc = false;
						body = wxString((const char *)&vals[valsta], wxConvISO8859_1);
					}
				} else {
					body = wxString((const char *)&vals[valsta], wxConvISO8859_1);
				}
				// trim
				body.Trim(true).Trim(false);
				// add line
				outlines.Add(body);
			}
			if (valpos >= valend) {
				// end of line
				phase = -1;
			} else {
				// continue
				phase = 1;
				valpos += lf_len;
			}
			valsta = valpos;
			lf_len = 0;
			break;
		}
	}

	return rc;
}
// テキストを出力
bool Parse::WriteText(wxArrayString &inlines, PsFileInfo &out_file)
{
	// UTF-8に変換するか
	bool utf8_mode = out_file.type.GetFlag(psUTF8);

	if (!out_file.file.IsOpened()) {
		return true;
	}

	if (!utf8_mode) {
		if (inlines.GetCount() > 0 && !inlines[0].IsEmpty()) {
			out_file.file.Write(nl_chr[mNewLineAscii], wxConvISO8859_1); // 1行目は必ず改行
		}

		for(size_t row = 0; row < inlines.GetCount(); row++) {
			out_file.file.Write(inlines[row], wxConvISO8859_1);	// 変換しない
			out_file.file.Write(nl_chr[mNewLineAscii], wxConvISO8859_1); // 改行
		}
	} else {
		if (out_file.type.GetFlag(psUTF8BOM)) {
			out_file.file.Write(BOM_CODE, 3); // BOM
		}

		for(size_t row = 0; row < inlines.GetCount(); row++) {
			out_file.file.Write(inlines[row], wxConvUTF8);	// UTF-8に変換して出力
			out_file.file.Write(nl_chr[mNewLineUtf8], wxConvISO8859_1);	// 改行
		}
	}
	return true;
}

// アスキー形式からUTF-8テキストに変換
bool Parse::ConvAsciiToUTF8(wxArrayString &inlines, PsFileType &out_file_type, wxArrayString *outlines, ParseResult *result)
{
	// 変換タイプ設定
	wxString out_char_type;
	if (out_file_type.GetFlag(psUTF8)) {
		// インターレースなどのUTF-8文字
		out_char_type = out_file_type.GetCharType();
	}

	wxString name = _("Ascii->UTF8");

	int error_count = 0;
	bool stopped = false;

	wxString body;

	size_t row = 0;

	for(row = 0; row < inlines.GetCount(); row++) {
		if (error_count > 100) {
			stopped = true;
			break;
		}
		error_count += ConvAsciiToUTF8OneLine(row, inlines[row], out_char_type, body, result);

		if (outlines) outlines->Add(body);
	}
	if (stopped) {
		if (result) {
			result->Add(row+1,1,GetLineNumber(inlines[row]),name,prErrStopInvalidToUTF8Code);
		}
		return false;
	}
	return true;
}

// アスキー形式1行からUTF-8テキストに変換
int Parse::ConvAsciiToUTF8OneLine(size_t row, const wxString &in_line, const wxString &out_type, wxString &out_line, ParseResult *result)
{
	// 変換タイプ設定
	bool through_mode = true;	// 半カナや記号をUTF-8に変換しないか

	if (!out_type.IsEmpty()) {
		// インターレースなどのUTF-8文字
		through_mode = false;
		mCharCodeTbl.FindSection(out_type);
	}

	wxString name = _("Ascii->UTF8");

	int error_count = 0;

	wxUint8 val[4];
	wxString body;
	wxString chrstr;
	CodeMapItem *item;

	out_line.Empty();
	for(size_t col = 0; col < in_line.Len();) {
		chrstr.Empty();
		// バイト列に変換
		AscStrToBytes(in_line.Mid(col), val, sizeof(val));
		if (!through_mode) {
			// キャラクターコードをUTF-8に変換
			item = mCharCodeTbl.FindByCode(val,_T("<"),false);
			if (item != NULL) {
				chrstr = item->GetStr();
				col += item->GetCodeLength();
			} else {
				if (val[0] >= 0x80) {
					if (result) {
						result->Add(row+1,col+1,GetLineNumber(in_line),name,prErrInvalidToUTF8Code);
					}
					error_count++;
				}
			}
		}
		if (chrstr.IsEmpty()) {
			chrstr = wxString(val, wxConvISO8859_1, 1);
			col++;
		}
		out_line += chrstr;
	}

	return error_count;
}

// UTF-8テキストからアスキー形式に変換
bool Parse::ConvUTF8ToAscii(PsFileType &in_file_type, wxArrayString &inlines, wxArrayString *outlines, ParseResult *result)
{
	// 変換タイプ設定
	wxString in_char_type;
	if (in_file_type.GetFlag(psUTF8)) {
		// インターレースなどのUTF-8文字
		in_char_type = in_file_type.GetCharType();
	}

	wxString name = _("UTF8->Ascii");

	wxString body;

	int error_count = 0;
	bool stopped = false;

	size_t row = 0;

	for(row = 0; row < inlines.GetCount(); row++) {
		if (error_count > 100) {
			stopped = true;
			break;
		}
		error_count += ConvUTF8ToAsciiOneLine(in_char_type, row, inlines[row], body, result);

		if (outlines) outlines->Add(body);
	}
	if (stopped) {
		if (result) {
			result->Add(row+1,1,GetLineNumber(inlines[row]),name,prErrStopInvalidUTF8Code);
		}
		return false;
	}

	return true;

}

// UTF-8テキスト１行からアスキー形式に変換
int Parse::ConvUTF8ToAsciiOneLine(const wxString &in_type, size_t row, const wxString &in_line, wxString &out_line, ParseResult *result)
{
	// 変換タイプ設定
	bool through_mode = true;	// UTF-8から半カナや記号に変換しないか
	if (!in_type.IsEmpty()) {
		// インターレースなどのUTF-8文字
		through_mode = false;
		mCharCodeTbl.FindSection(in_type);
	}

	wxString name = _("UTF8->Ascii");

	wxCharBuffer vals;
	size_t vallen;
	wxString chrstr;
	long     chr;
	CodeMapItem *item;

	int error_count = 0;

	size_t col = 0;

	out_line.Empty();
	vals = in_line.mb_str(wxConvUTF8);
	vallen = strlen(vals);
	for(col = 0; col < vallen;) {
		if (vals[col] < 0) { // upper 0x80
			// UTF-8 code?
			if (vals[col] == 0xef && vals[col+1] == 0xbb && vals[col+2] == 0xbf) {
				// BOM?
				col += 3;
			} else {
				chr = 0;
				chrstr.Empty();
				if (!through_mode) {
					// UTF-8をキャラクターコードに変換
					item = mCharCodeTbl.FindByBytes((const wxUint8 *)((const char *)vals+col),_T(">"),false);
					if (item != NULL) {
						chrstr = wxString::From8BitData((const char *)item->GetCode(), item->GetCodeLength());
						col += item->GetBytesLength();
					} else {
						// cannot convert
						if (result) {
							result->Add(row+1,col+1,GetLineNumber(in_line),name,prErrInvalidUTF8Code);
						}
						error_count++;
					}
				}
				if (chrstr.IsEmpty()) {
					chrstr = wxString::From8BitData((const char *)vals + col, 1);
					col++;
				}
				out_line += chrstr;
			}
		} else {
			// ascii char
			chrstr = wxString::From8BitData((const char *)vals + col, 1);
			col++;
			out_line += chrstr;
		}
	}

	return error_count;
}

// 文字コード変換テーブルの読み込み
PsErrType Parse::LoadCharCodeTable()
{
	wxTextFile file;
	wxFileName filename(mAppPath + DATA_DIR, CHAR_CODE_TABLE);
	wxString path = filename.GetFullPath();

	if (!file.Open(path)) {
		mErrInfo.SetInfo(__LINE__, psError, psErrCannotOpen, psInfoFileInApp, CHAR_CODE_TABLE);
		mErrInfo.ShowMsgBox();
		return psError;
	}

	wxString line;
	for ( line = file.GetFirstLine(); !file.Eof(); line = file.GetNextLine() ) {
		// skip if comment line or empty
		line.Trim(false).Trim(true);
		if ( line.IsEmpty() || line[0] == wxChar('#') ) {
			continue;
		}

		// charset type code
		wxRegEx re(_T("^\\[(.+)\\]$"));
		if (re.Matches(line)) {
			wxString mstr = re.GetMatch(line, 1);
			mCharCodeTbl.AddSection(mstr);
			continue;
		}

		int end_pos = line.Find(wxChar(','));
		if (end_pos == wxNOT_FOUND) {
			continue;
		}
		wxString new_code_str = line.Left(end_pos);
		new_code_str.Trim(true).Trim(false);
		wxUint8 new_code[4];
		size_t new_code_len = HexStrToBytes(new_code_str, new_code, sizeof(new_code));
		line = line.Mid(end_pos+1);

		end_pos = line.Find(wxChar(','));
		if (end_pos == wxNOT_FOUND) {
			end_pos = line.Len();
		}
		wxString new_str = line.Left(end_pos);
		int sep_st_pos = new_str.Find(wxChar('\''), false);
		int sep_ed_pos = new_str.Find(wxChar('\''), true);
		if (sep_st_pos != wxNOT_FOUND && sep_ed_pos != wxNOT_FOUND) {
			new_str = new_str.SubString(sep_st_pos+1, sep_ed_pos-1);
		} else {
			new_str.Empty();
		}
		line = line.Mid(end_pos+1);

		wxString new_attr = line;
		new_attr.Trim(true).Trim(false);

		mCharCodeTbl.AddItem(new_code, new_code_len, new_str, new_attr);
	}

	file.Close();
	return psOK;
}

// BASICコード変換テーブルの読み込み
PsErrType Parse::LoadBasicCodeTable()
{
	wxTextFile file;
	wxFileName filename(mAppPath + DATA_DIR, BASIC_CODE_TABLE);
	wxString path = filename.GetFullPath();

	if (!file.Open(path)) {
		mErrInfo.SetInfo(__LINE__, psError, psErrCannotOpen, psInfoFileInApp, BASIC_CODE_TABLE);
		mErrInfo.ShowMsgBox();
		return psError;
	}

	wxString line;
	for ( line = file.GetFirstLine(); !file.Eof(); line = file.GetNextLine() ) {
		// skip if comment line or empty
		line.Trim(false).Trim(true);
		if ( line.IsEmpty() || line[0] == wxChar('#') ) {
			continue;
		}

		// basic type
		wxRegEx re(_T("^\\[(.*)\\]$"));
		if (re.Matches(line)) {
			wxString mstr = re.GetMatch(line, 1);
			mBasicCodeTbl.AddSection(mstr);
			continue;
		}

		int end_pos = line.Find(wxChar(','));
		if (end_pos == wxNOT_FOUND) {
			continue;
		}
		wxString new_code_str = line.Left(end_pos);
		new_code_str.Trim(true).Trim(false);
		wxUint8 new_code[4];
		size_t new_code_len = HexStrToBytes(new_code_str, new_code, sizeof(new_code));
		line = line.Mid(end_pos+1);

		end_pos = line.Find(wxChar(','));
		if (end_pos == wxNOT_FOUND) {
			end_pos = line.Len();
		}
		wxString new_str = line.Left(end_pos);
		new_str.Trim(true).Trim(false);
		line = line.Mid(end_pos+1);

		wxString new_attr = line;
		new_attr.Trim(true).Trim(false);

		mBasicCodeTbl.AddItem(new_code, new_code_len, new_str, new_attr);
	}

	file.Close();
	return psOK;
}

// 文字種類を返す
void Parse::GetCharTypes(wxArrayString &char_types)
{
	mCharCodeTbl.GetAllSectionNames(char_types);
}
wxString &Parse::GetCharType(size_t index)
{
	return mCharCodeTbl.GetSectionName(index);
}

// BASIC種類を返す
void Parse::GetBasicTypes(wxArrayString &basic_types)
{
	mBasicCodeTbl.GetAllSectionNames(basic_types);
}

// BASIC種類をセット
//void Parse::SetBasicType(const wxString &basic_type)
//{
//	mInFile.type.SetBasicType(basic_type);
//	mOutFile.type.SetBasicType(basic_type);
//	if (!basic_type.IsEmpty()) {
//		mInFile.type.SetFlag(psExtendBasic, true);
//		mOutFile.type.SetFlag(psExtendBasic, true);
//	}
//}

// 入力ファイルのBASIC種類を返す
wxString &Parse::GetOpenedBasicType()
{
	return mInFile.type.GetBasicType();
}

// 画面表示用データを返す
wxString &Parse::GetParsedData(wxArrayString &lines)
{
	for(size_t i=0; i<mParsedData.GetCount(); i++) {
		lines.Add(mParsedData[i]);
	}
	return mParsedType.GetCharType();
}

// 画面表示用データ種類
//int Parse::GetParsedDataType()
//{
//	return mParsedType.flags;
//}
PsFileType *Parse::GetParsedDataTypePtr()
{
	return &mParsedType;
}

// スタートアドレスのリストへのポインタを返す
int *Parse::GetStartAddrsPtr(size_t *count)
{
	if (count) *count = sizeof(iStartAddr) / sizeof(int);
	return iStartAddr;
}

// 行番号を返す
long Parse::GetLineNumber(const wxString &line, size_t *next_pos)
{
	long num = 0;

	// get line number
	int pos = line.Find(wxChar(' '));
	if (pos != wxNOT_FOUND) {
		line.Left(pos).ToLong(&num);
		if (next_pos) *next_pos = pos + 1;
	} else {
		wxRegEx re(_T("^\\d+"));
		if (re.Matches(line)) {
			wxString sub = re.GetMatch(line);
			sub.ToLong(&num);
			if (next_pos) *next_pos = sub.Len();
		}
	}
	return num;
}

// レポート
void Parse::Report(ParseResult &result, wxArrayString &line)
{
	wxArrayString arr;
	result.Report(arr);
	if (arr.GetCount() == 0) return;

	line.Add(_("----- Result Report -----"));
	for(size_t i=0; i<arr.GetCount(); i++) {
		line.Add(arr[i]);
	}
	line.Add(_("-------------------------"));
	line.Add(_T(""));
}

// 16進文字列をバイト列(可変)に変換(big endien)
/// @param[in]  src     16進文字列
/// @param[out] dst     バイト列
/// @param[in]  dst_len dstの長さ
/// @return     書き込んだ長さ
int Parse::HexStrToBytes(const wxString &src, wxUint8 *dst, size_t dst_len)
{
	unsigned long val;

	if (src.ToULong(&val, 16) != true) {
		return 0;
	}

	size_t len;
	for(len = 3; len > 0; len--) {
		if (val >= (unsigned long)(1 << (len << 3))) {
			break;
		}
	}

	memset(dst, 0, dst_len);
	for(size_t i=0; i < dst_len && i <= len; i++) {
		dst[i] = (val >> ((len - i) << 3)) & 0xff;
	}

	return (int)(len + 1);
}

// アスキー文字列をバイト列に変換
/// @param[in]  src     アスキー文字列
/// @param[out] dst     バイト列
/// @param[in]  dst_len dstの長さ
/// @return     書き込んだ長さ
int Parse::AscStrToBytes(const wxString &src, wxUint8 *dst, size_t dst_len)
{
	size_t len;
	wxChar c;

	memset(dst, 0, dst_len);
	for(len=0; len < dst_len && len < src.Len(); len++) {
		c = src[len];
		dst[len] = (wxUint8)c;
	}

	return (int)len;
}

// 整数バイト列を数値に変換
/// @param[in]  src     整数バイト列
/// @param[in]  src_len srcの長さ
/// @return     数値
long Parse::BytesToLong(const wxUint8 *src, size_t src_len)
{
	long num = 0;

	src_len--;
	if (src_len > 3) src_len = 3;

	if (mBigEndian) {
		for(int i=0; i<=(int)src_len; i++) {
			num <<= 8;
			num += src[i];
		}
	} else {
		for(int i=(int)src_len; i>=0; i--) {
			num <<= 8;
			num += src[i];
		}
	}
	return num;
}

// 整数バイト列を文字列に変換
/// @param[in]  src     整数バイト列
/// @param[in]  src_len srcの長さ
/// @param[out] dst     文字列
/// @return     true
bool Parse::BytesToStr(const wxUint8 *src, size_t src_len, wxString &dst)
{
	long num = BytesToLong(src, src_len);
	dst = wxString::Format("%ld", num);
	return true;
}

// 数値を整数バイト列に変換
/// @param[in]  src     数値
/// @param[out] dst     整数バイト列
/// @param[in]  dst_len dstの長さ
/// @return     true
bool Parse::LongToBytes(long src, wxUint8 *dst, size_t dst_len)
{
	dst_len--;
	if (dst_len > 3) dst_len = 3;

	if (mBigEndian) {
		for(int i=(int)dst_len; i>=0; i--) {
			dst[i] = (src & 0xff);
			src >>= 8;
		}
	} else {
		for(int i=0; i<=(int)dst_len; i++) {
			dst[i] = (src & 0xff);
			src >>= 8;
		}
	}
	return true;
}

