﻿/// @file parse.h
///
/// @brief パーサー
///
#ifndef _PARSE_H_
#define _PARSE_H_

#include "common.h"
#include <wx/wx.h>
#include <wx/file.h>
#include <wx/datstrm.h>
#include <wx/dynarray.h>
#include "errorinfo.h"
#include "maptable.h"
#include "fileinfo.h"
#include "parseresult.h"

/// パーサークラス
class Parse
{
private:
	wxString mAppPath;
	bool mBigEndian;

	PsErrInfo mErrInfo;	// エラー情報

	PsFileInfo mInFile;		// 入力ファイル情報
	PsFileInfo mOutFile;		// 出力ファイル情報

	int mNextAddress;		// 次アドレス

	PsFileType     mParsedType;
	wxArrayString mParsedData;	// 画面表示用データバッファ

	wxString mAsciiNewLine;	// 改行コード
	wxString mUtf8NewLine;	// 改行コード

	CodeMapTable mCharCodeTbl;	// 文字コード変換テーブル
	CodeMapTable mBasicCodeTbl;	// BASICコード変換テーブル

	// パラメータ
	int mNewLineAscii;
	int mNewLineUtf8;
	bool mIncludeBOM;
	int  mStartAddr;

	/// 入力データのフォーマットチェック
	bool CheckDataFormat(PsFileInfo &in_file_info);
	/// 中間言語形式データのフォーマットチェック
	bool CheckBinaryDataFormat(PsFileInfo &in_file_info);
	/// アスキー形式データのフォーマットチェック
	bool CheckAsciiDataFormat(PsFileInfo &in_file_info);
	/// 中間言語形式データの読み直し
	bool ReloadBinaryData(PsFileInfo &in_file_info, PsFileType &out_type);
	/// アスキー形式データの読み直し
	bool ReloadAsciiData(PsFileInfo &in_file_info, PsFileType &out_type);
	/// 中間言語からアスキー形式テキストに変換
	bool ReadBinaryToAscii(PsFileInfo &in_file, PsFileType &out_file_type, wxArrayString &outlines, ParseResult *result = NULL);
	/// アスキー形式から中間言語に変換
	bool ParseAsciiToBinary(PsFileType &in_file_type, wxArrayString &inlines, PsFileInfo &out_file, ParseResult *result = NULL);
	/// アスキー形式1行を中間言語に変換
	bool ParseAsciiToBinaryOneLine(PsFileType &in_file_type, wxString &in_data, PsFileInfo &out_file, int row = 0, ParseResult *result = NULL);
	/// アスキー形式テキストを読む
	bool ReadAsciiText(PsFileInfo &in_file, wxArrayString &outlines, bool to_utf8 = false);
	/// テキストを出力
	bool WriteText(wxArrayString &inlines, PsFileInfo &out_file);
	/// アスキー形式からUTF-8テキストに変換
	bool ConvAsciiToUTF8(wxArrayString &inlines, PsFileType &out_file_type, wxArrayString *outlines, ParseResult *result = NULL);
	/// UTF-8テキストからアスキー形式に変換
	bool ConvUTF8ToAscii(PsFileType &in_file_type, wxArrayString &inlines, wxArrayString *outlines, ParseResult *result = NULL);
	/// 文字コード変換テーブルの読み込み
	PsErrType LoadCharCodeTable();
	/// BASICコード変換テーブルの読み込み
	PsErrType LoadBasicCodeTable();
	/// 16進文字列をバイト列に変換
	int  HexStrToBytes(const wxString &src, wxUint8 *dst, size_t dst_len);
	/// アスキー文字列をバイト列に変換
	int  AscStrToBytes(const wxString &src, wxUint8 *dst, size_t dst_len);
	/// 整数バイト列を数値に変換
	long BytesToLong(const wxUint8 *src, size_t src_len);
	/// 整数バイト列を数値に変換
	bool BytesToStr(const wxUint8 *src, size_t src_len, wxString &dst);
	/// 数値を整数バイト列に変換
	bool LongToBytes(long src, wxUint8 *dst, size_t dst_len);

public:
	Parse(const wxString &path);
	~Parse();
	bool Init();
	/// 指定したファイルを開く
	bool OpenDataFile(const wxString &in_file_name);
	/// ファイルを閉じる
	void CloseDataFile();
	/// ファイル開いているか
	bool IsOpenedDataFile();
	/// 開いているファイル種類
	PsFileType *GetOpenedDataTypePtr();
	/// 開いているファイル情報
	PsFileInfo *GetOpenedDataInfoPtr();
	/// ファイル名
	wxString GetFileNameBase();
	/// 出力ファイルを開く
	bool OpenOutFile(const wxString &out_file_name, PsFileType &file_type);
	/// 出力ファイルを閉じる
	void CloseOutFile();
	/// エクスポート
	bool ExportData();
	/// 画面表示用データを返す
	wxString &GetParsedData(wxArrayString &lines);
	/// 画面表示用データ種類
	PsFileType *GetParsedDataTypePtr();
	/// 文字種類を返す
	void GetCharTypes(wxArrayString &char_types);
	wxString &GetCharType(size_t index);
	/// BASIC種類を返す
	void GetBasicTypes(wxArrayString &basic_types);
//	/// BASIC種類をセット
//	void SetBasicType(const wxString &basic_type);
	/// 入力ファイルのBASIC種類を返す
	wxString &GetOpenedBasicType();
	/// 入力ファイルの形式を変更して読み直す
	bool ReloadOpendAsciiData(int type, int mask, const wxString &char_type);
	/// 表示用データの形式を変更して読み直す
	bool ReloadParsedData(int type, int mask, const wxString &char_type);
	/// 中間言語形式データの読み直し
	bool ReloadOpenedBinaryData(int type, int mask, const wxString &basic_type);
	/// スタートアドレスのリストへのポインタを返す
	int *GetStartAddrsPtr(size_t *count);
	/// 行番号を返す
	long GetLineNumber(const wxString &line, size_t *next_pos = NULL);
	/// レポート
	void Report(ParseResult &result, wxArrayString &line);
	/// アスキー形式1行からUTF-8テキストに変換
	int  ConvAsciiToUTF8OneLine(size_t row, const wxString &in_line, const wxString &out_type, wxString &out_line, ParseResult *result = NULL);
	// UTF-8テキスト1行からアスキー形式に変換
	int  ConvUTF8ToAsciiOneLine(const wxString &in_type, size_t row, const wxString &in_line, wxString &out_line, ParseResult *result = NULL);

	/// パラメータ
	void SetNewLineAscii(int val) { mNewLineAscii = val; }
	void SetNewLineUtf8(int val) { mNewLineUtf8 = val; }
	void SetIncludeBOM(bool val) { mIncludeBOM = val; }
	void SetStartAddr(int val) { mStartAddr = val; }
	int GetNewLineAscii() { return mNewLineAscii; }
	int GetNewLineUtf8() { return mNewLineUtf8; }
	bool GetIncludeBOM() { return mIncludeBOM; }
	int GetStartAddr() { return mStartAddr; }

};

#endif /* _PARSE_H_ */
